package de.unibi.citec.clf.bonsai.util;

/*
 * #LICENSE%
 * BonSAI [Core|Base]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;
import de.unibi.airobots.btl.units.UnitConverter;

/**
 * Utility class with methods to convert between several coordinate systems.
 */
public class CoordinateSystemConverter {

	/**
	 * Generate global carthesian coordinates suitable as a navigation goal for
	 * the NavigationActuator from local polar coordinate system coordinates
	 * given as distance and angle. This method requires the current global
	 * position (OdometryData) of the robot obtained from the OdometrySensor.
	 * 
	 * 
	 * @param positionData
	 *            global position of the robot as recently obtained from an
	 *            OdometrySensor
	 * @param angle
	 *            angle of the polar coordinates. An angle=0 corresponds to a
	 *            position directly in front of the robot. angles are defined
	 *            counter-clockwise.
	 * @param dist
	 *            distance of polar coordinates (always >0).
	 * @param angleUnit
	 *            unit of angle.
	 * @param lengthUnit
	 *            unit of dist.
	 * @return the global carthesian position in metres and the corresponding
	 *         orientation of the robot as NavigationGoalData, which can
	 *         directly by used for the NavigationActuator.
	 */
	public static NavigationGoalData polar2NavigationGoalData(
			PositionData positionData, double angle, double dist,
			AngleUnit angleUnit, LengthUnit lengthUnit) {
		NavigationGoalData cart = new NavigationGoalData();
		double localAngle = UnitConverter.convert(angle, angleUnit,
				AngleUnit.RADIAN) + positionData.getYaw(AngleUnit.RADIAN);

		if (localAngle > 2 * Math.PI) {
			localAngle = localAngle - 2 * Math.PI;
		}
		if (localAngle < 0) {
			localAngle = localAngle + 2 * Math.PI;
		}
		cart.setX(positionData.getX(LengthUnit.METER) + Math.cos(localAngle)
				* UnitConverter.convert(dist, lengthUnit, LengthUnit.METER),
				LengthUnit.METER);

		cart.setY(positionData.getY(LengthUnit.METER) + Math.sin(localAngle)
				* UnitConverter.convert(dist, lengthUnit, LengthUnit.METER),
				LengthUnit.METER);
		cart.setYaw(localAngle, AngleUnit.RADIAN);
		return cart;
	}

	/**
	 * Generate local carthesian coordinates suitable as a navigation goal for
	 * the NavigationActuator from local polar coordinate system coordinates
	 * given as distance and angle. This method requires the current global
	 * position (OdometryData) of the robot obtained from the OdometrySensor.
	 * 
	 * 
	 * @param positionData
	 *            global position of the robot as recently obtained from an
	 *            OdometrySensor
	 * @param angle
	 *            angle of the polar coordinates. An angle=0 corresponds to a
	 *            position directly in front of the robot. angles are defined
	 *            counter-clockwise.
	 * @param dist
	 *            distance of polar coordinates (always >0).
	 * @param angleUnit
	 *            unit of angle.
	 * @param lengthUnit
	 *            unit of dist.
	 * @return the global carthesian position in metres and the corresponding
	 *         orientation of the robot as NavigationGoalData, which can
	 *         directly by used for the NavigationActuator.
	 */
	public static NavigationGoalData polar2LocalNavigationGoalData(
			double angle, double dist, AngleUnit angleUnit,
			LengthUnit lengthUnit) {

		PositionData localPosition = new PositionData();
		localPosition.setX(0.0, lengthUnit);
		localPosition.setY(0.0, lengthUnit);
		localPosition.setYaw(0.0, angleUnit);

		NavigationGoalData cart = polar2NavigationGoalData(localPosition,
				angle, dist, angleUnit, lengthUnit);
		cart.setType(NavigationGoalData.GoalType.LOCAL);

		return cart;
	}

	public static double positionData2Angle(PositionData ownPosition,
			PositionData objectPosition, AngleUnit angleUnit) {

		double angle = Math.atan2(
				objectPosition.getY(LengthUnit.METER)
						- ownPosition.getY(LengthUnit.METER),
				objectPosition.getX(LengthUnit.METER)
						- ownPosition.getX(LengthUnit.METER));

		angle += ownPosition.getYaw(AngleUnit.RADIAN);

		if (angle >= Math.PI) {
			angle -= 2 * Math.PI;
		}
		if (angle < -Math.PI) {
			angle += 2 * Math.PI;
		}

		return UnitConverter.convert(angle, AngleUnit.RADIAN, angleUnit);
	}

	public static double positionDistance(PositionData ownPosition,
			PositionData objectPosition, LengthUnit lUnit) {

		double x = ownPosition.getX(lUnit) - objectPosition.getX(lUnit);

		double y = ownPosition.getY(lUnit) - objectPosition.getY(lUnit);

		double dist = Math.sqrt(Math.pow(x, 2) + Math.pow(y, 2));

		return dist;
	}

	public static PositionData localToGlobal(PositionData src,
			PositionData currentPosition) {

		double x = src.getX(LengthUnit.METER);
		double y = src.getY(LengthUnit.METER);
		double yaw = src.getYaw(AngleUnit.RADIAN);

		// rotate the src point into the global cs
		double cs = Math.cos(-currentPosition.getYaw(AngleUnit.RADIAN));
		double sn = Math.sin(-currentPosition.getYaw(AngleUnit.RADIAN));
		double px = x * cs - y * sn;
		double py = x * sn + y * cs;

		// translate into the global cs
		double gx = currentPosition.getX(LengthUnit.METER) + px;
		double gy = currentPosition.getY(LengthUnit.METER) + py;
		double gyaw = currentPosition.getYaw(AngleUnit.RADIAN) + yaw;

		PositionData tgt = new PositionData(gx, gy, gyaw, src.getTimestamp(),
				LengthUnit.METER, AngleUnit.RADIAN);
		return tgt;
	}
}
