package de.unibi.citec.clf.bonsai.core;

/*
 * #LICENSE%
 * BonSAI [Core|Base]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * {@link CoreObjectFactory} used for unit tests.
 * 
 * @author jwienke
 */
public class TestFactory implements CoreObjectFactory {

	public static final String FAIL_KEY = "fail";

	private Map<String, Class<?>> knownSensorKeysToType = new HashMap<String, Class<?>>();
	private Map<String, Class<? extends List<?>>> knownListSensorKeysToType = new HashMap<String, Class<? extends List<?>>>();
	private Map<String, Class<? extends Actuator>> knownActuatorKeysToImplClass = new HashMap<String, Class<? extends Actuator>>();
	private Map<String, Class<? extends WorkingMemory>> knownMemoryKeysToImplClass = new HashMap<String, Class<? extends WorkingMemory>>();

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void configureSensors(Set<SensorToConfigure> sensors)
			throws IllegalArgumentException, CoreObjectCreationException {

		for (SensorToConfigure s : sensors) {

			if (s.getKey() == null) {
				throw new CoreObjectCreationException("empty key");
			}
			if (s.getSensorClass() == null) {
				throw new CoreObjectCreationException("empty class");
			}

			if (knownSensorKeysToType.containsKey(s.getKey())) {
				throw new CoreObjectCreationException("duplicate key");
			}

			knownSensorKeysToType.put(s.getKey(), s.getDataTypeClass());

			if (s.isListSensor()) {
				knownListSensorKeysToType.put(s.getKey(), s.getListTypeClass());
			}
		}

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <T> Sensor<T> createSensor(String key, Class<T> dataType)
			throws IllegalArgumentException, CoreObjectCreationException {
		if (knownSensorKeysToType.containsKey(key)
				&& knownSensorKeysToType.get(key).equals(dataType)) {
			return new TestSensor<T>(dataType);
		} else {
			throw new IllegalArgumentException("Unknown key-type combination.");
		}
	}

	@SuppressWarnings("unchecked")
	public static <S extends List<T>, T> Class<S> createListClassOfType(
			Class<S> list, Class<T> type) throws InstantiationException,
			IllegalAccessException {

		return (Class<S>) list.newInstance().getClass();
	}

	@Override
	public <S extends List<T>, T> Sensor<S> createSensor(String key,
			Class<S> listType, Class<T> dataType)
			throws IllegalArgumentException, CoreObjectCreationException {

		try {
			if (knownListSensorKeysToType.containsKey(key)
					&& knownSensorKeysToType.get(key).equals(dataType)
					&& knownListSensorKeysToType.get(key).equals(listType)) {
				Class<S> clazz = createListClassOfType(listType, dataType);
				return new TestSensor<S>(clazz);
			} else {
				throw new IllegalArgumentException(
						"Unknown key-type combination.");
			}
		} catch (IllegalAccessException e) {
			throw new CoreObjectCreationException(e);
		} catch (InstantiationException e) {
			throw new CoreObjectCreationException(e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initialize(Map<String, String> options)
			throws IllegalArgumentException {
		if (options.containsKey(FAIL_KEY)) {
			throw new IllegalArgumentException();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean canCreateSensor(String key, Class<?> dataType) {
		return knownSensorKeysToType.containsKey(key)
				&& knownSensorKeysToType.get(key).equals(dataType);
	}

	@Override
	public boolean canCreateSensor(String key,
			Class<? extends List<?>> listType, Class<?> dataType) {
		return knownListSensorKeysToType.containsKey(key)
				&& knownSensorKeysToType.get(key).equals(dataType)
				&& knownListSensorKeysToType.get(key).equals(listType);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void configureActuators(Set<ActuatorToConfigure> actuators)
			throws IllegalArgumentException, CoreObjectCreationException {

		for (ActuatorToConfigure a : actuators) {

			if (a.getKey() == null) {
				throw new CoreObjectCreationException("empty key");
			}
			if (a.getActuatorClass() == null) {
				throw new CoreObjectCreationException("empty class");
			}

			if (knownActuatorKeysToImplClass.containsKey(a.getKey())) {
				throw new CoreObjectCreationException("duplicate key");
			}

			knownActuatorKeysToImplClass.put(a.getKey(), a.getActuatorClass());

		}

	}

	@Override
	public void configureWorkingMemories(Set<WorkingMemoryToConfigure> memories)
			throws IllegalArgumentException, CoreObjectCreationException {
		
		for (WorkingMemoryToConfigure a : memories) {

			if (a.getKey() == null) {
				throw new CoreObjectCreationException("empty key");
			}
			if (a.getMemoryClass() == null) {
				throw new CoreObjectCreationException("empty class");
			}

			if (knownMemoryKeysToImplClass.containsKey(a.getKey())) {
				throw new CoreObjectCreationException("duplicate key");
			}

			knownMemoryKeysToImplClass.put(a.getKey(), a.getMemoryClass());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean canCreateActuator(String key,
			Class<? extends Actuator> actuatorClass) {
		return knownActuatorKeysToImplClass.containsKey(key)
				&& actuatorClass.isAssignableFrom(knownActuatorKeysToImplClass
						.get(key));
	}

	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	@Override
	public <T extends Actuator> T createActuator(String key,
			Class<T> actuatorClass) throws IllegalArgumentException,
			CoreObjectCreationException {

		if (!canCreateActuator(key, actuatorClass)) {
			throw new IllegalArgumentException(
					"Cannot create an actuator with key '" + key + "' of type "
							+ actuatorClass);
		}

		try {
			return (T) knownActuatorKeysToImplClass.get(key).newInstance();
		} catch (InstantiationException e) {
			throw new CoreObjectCreationException(e);
		} catch (IllegalAccessException e) {
			throw new CoreObjectCreationException(e);
		}

	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends WorkingMemory> T createWorkingMemory(String key)
			throws IllegalArgumentException, CoreObjectCreationException {
		
		if (!canCreateWorkingMemory(key)) {
			throw new IllegalArgumentException(
					"Cannot create an working memory with key '" + key + "'");
		}

		try {
			return (T) knownMemoryKeysToImplClass.get(key).newInstance();
		} catch (InstantiationException e) {
			throw new CoreObjectCreationException(e);
		} catch (IllegalAccessException e) {
			throw new CoreObjectCreationException(e);
		}
	}

	@Override
	public boolean canCreateWorkingMemory(String key) {
		return knownMemoryKeysToImplClass.containsKey(key)
				&& WorkingMemory.class
						.isAssignableFrom(knownMemoryKeysToImplClass.get(key));
	}

}
