package de.unibi.citec.clf.bonsai.xcf;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.Actuator;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory;
import de.unibi.citec.clf.bonsai.core.Sensor;
import de.unibi.citec.clf.bonsai.core.WorkingMemory;
import de.unibi.citec.clf.bonsai.util.reflection.JarServiceDiscovery;
import de.unibi.citec.clf.bonsai.util.reflection.ServiceDiscovery;
import de.unibi.citec.clf.bonsai.xcf.configuration.ConfiguredXcfActuator;
import de.unibi.citec.clf.bonsai.xcf.configuration.ConfiguredXcfCoreObject;
import de.unibi.citec.clf.bonsai.xcf.configuration.ConfiguredXcfSensor;
import de.unibi.citec.clf.bonsai.xcf.configuration.ConfiguredXcfSensorListable;
import de.unibi.citec.clf.bonsai.xcf.configuration.ConfiguredXcfWorkingMemory;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import net.sf.xcf.InitializeException;
import net.sf.xcf.XcfManager;
import org.apache.log4j.Logger;

/**
 * Factory for Bonsai core objects that use the XCF middleware.
 *
 * <p> This factory uses a {@link ServiceDiscovery} for finding sensor and
 * actuator implementations. The default implementation is
 * {@link JarServiceDiscovery}. To add new actuators or sensors to the list of
 * known ones, register them in
 * <code>META-INF/services</code> in your jar file and load this jar file in the
 * class path. A more detailed description is given at the JAR specification
 * linked at {@link JarServiceDiscovery}. </p>
 *
 * Options supported by this factory: <ul>
 * <li>{@value #OPTION_ERROR_ON_INITIAL_SUBSCRIPTION} (boolean): throw
 * exceptions if the initial connection to a publisher fails (default =
 * {@link ReconnectableManager#DEFAULT_ERROR_ON_TYPE_INITIALIZATION})</li>
 * <li>{@value #OPTION_SUBSCRIBER_CHECK_INTERVAL} (long): interval used to check
 * that subscribers are alive in ms. (default =
 * {@link ReconnectableManager#DEFAULT_CHECK_INTERVAL_IN_MS})</li>
 * <li>{@value #OPTION_REMOTE_SERVER_CHECK_INTERVAL} (long): interval used to
 * check that remote servers are alive in ms. (default =
 * {@link ReconnectableManager#DEFAULT_CHECK_INTERVAL_IN_MS})</li> </ul>
 *
 * TODO: <ul> <li>Implement active memory reconnection. Currently this is not
 * possible because of <a
 * href="https://code.ai.techfak.uni-bielefeld.de/trac/xcf/ticket/337">this XCF
 * ticket</a></li> <li>Add a user option to specify the {@link ServiceDiscovery}
 * implementation to use for finding sensors and actuators.</li> </ul>
 *
 * @author jwienke
 */
public class XcfFactory implements CoreObjectFactory {

    /**
     * Option to report errors on initial connections to publishers.
     */
    public static final String OPTION_ERROR_ON_INITIAL_SUBSCRIPTION = "errorOnInitialSubscription";
    /**
     * Option defining the interval used to check that subscribers are alive.
     */
    public static final String OPTION_SUBSCRIBER_CHECK_INTERVAL = "subscriberCheckInterval";
    /**
     * Option defining the interval used to check that remote servers are alive.
     */
    public static final String OPTION_REMOTE_SERVER_CHECK_INTERVAL = "remoteServerCheckInterval";
    private Logger logger = Logger.getLogger(getClass());
    protected XcfManager xcfManager = null;
    /**
     * Name of the static method in {@link ConfiguredXcfCoreObject}s for
     * verifying that a configuration object suites a requested sensor or
     * actuator configuration.
     */
    private static final String IS_SUITABLE_METHOD_NAME = "isSuitableFor";
    /**
     * The list of know configuration objects used to configure and create new
     * actuator instances.
     */
    protected Set<Class<? extends ConfiguredXcfActuator>> knownActuators = new HashSet<Class<? extends ConfiguredXcfActuator>>();
    /**
     * The list of know configuration objects used to configure and create new
     * actuator instances.
     */
    protected Set<Class<? extends ConfiguredXcfWorkingMemory>> knownWorkingMemories = new HashSet<Class<? extends ConfiguredXcfWorkingMemory>>();
    /**
     * The list of know configuration objects used to configure and create new
     * sensor instances.
     */
    @SuppressWarnings("rawtypes")
    protected Set<Class<? extends ConfiguredXcfSensor>> knownSensors = new HashSet<Class<? extends ConfiguredXcfSensor>>();
    /**
     * The list of know configuration objects used to configure and create new
     * sensor instances.
     */
    @SuppressWarnings("rawtypes")
    protected Set<Class<? extends ConfiguredXcfSensorListable>> knownListSensors = new HashSet<Class<? extends ConfiguredXcfSensorListable>>();
    /**
     * Map of configured sensor keys mapped to their configuration object.
     */
    protected Map<String, ConfiguredXcfSensor<?>> configuredSensorsByKey = new HashMap<String, ConfiguredXcfSensor<?>>();
    /**
     * Map of configured sensor keys mapped to their configuration object.
     */
    protected Map<String, ConfiguredXcfSensorListable<? extends List<?>, ?>> configuredListSensorsByKey = new HashMap<String, ConfiguredXcfSensorListable<? extends List<?>, ?>>();
    /**
     * Map of configured actuator keys mapped to their configuration object.
     */
    protected Map<String, ConfiguredXcfActuator> configuredActuatorsByKey = new HashMap<String, ConfiguredXcfActuator>();
    protected Map<String, ConfiguredXcfWorkingMemory> configuredMemoriesByKey = new HashMap<String, ConfiguredXcfWorkingMemory>();
    protected SubscriberReconnectableManager subscriberManager = null;
    protected RemoteServerReconnectableManager remoteServerManager = null;
    /**
     * Service discovery mechanism used for finding actuator and sensor
     * implementations.
     */
    protected ServiceDiscovery serviceDiscovery = new JarServiceDiscovery();

    /**
     * Constructor.
     */
    public XcfFactory() {
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canCreateActuator(String key,
            Class<? extends Actuator> actuatorClass) {
        // first check that there was a actuator with this key
        if (!configuredActuatorsByKey.containsKey(key)) {
            return false;
        }

        return actuatorClass.isAssignableFrom(configuredActuatorsByKey.get(key)
                .getActuatorClass());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canCreateSensor(String key, Class<?> dataType) {

        // first check that there was a sensor with this key
        if (!configuredSensorsByKey.containsKey(key)) {
            return false;
        }

        return dataType.isAssignableFrom(configuredSensorsByKey.get(key)
                .getDataType());

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canCreateSensor(String key,
            Class<? extends List<?>> listType, Class<?> dataType) {

        // first check that there was a sensor with this key
        if (!configuredListSensorsByKey.containsKey(key)) {
            return false;
        }

        return dataType.isAssignableFrom(configuredListSensorsByKey.get(key)
                .getDataType());

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void configureActuators(Set<ActuatorToConfigure> actuators)
            throws IllegalArgumentException, CoreObjectCreationException {

        logger.info("Configuring actuators: " + actuators);

        actuatorLoop:
        for (ActuatorToConfigure actuator : actuators) {
            logger.debug("Processing actuator to configure: " + actuator);

            // check that there is no other actuator with the requested key
            if (configuredActuatorsByKey.containsKey(actuator.getKey())) {
                throw new IllegalArgumentException("Duplicate actuator key '"
                        + actuator.getKey() + "' requested.");
            }

            // find the actuator configuration that can handle this requested
            // actuator
            for (Class<? extends ConfiguredXcfActuator> configurationClass : knownActuators) {

                logger.debug("Checking if configuration class "
                        + configurationClass + " can create actuator "
                        + actuator);

                try {

                    Method isSuitableForMethod = configurationClass.getMethod(
                            IS_SUITABLE_METHOD_NAME, ActuatorToConfigure.class);
                    Boolean isSuitable = (Boolean) isSuitableForMethod.invoke(
                            null, actuator);

                    // if this configuration object is suitable, create an
                    // instance and initialize it
                    if (isSuitable) {

                        logger.debug("Configuration class "
                                + configurationClass + " can create actuator "
                                + actuator);

                        ConfiguredXcfActuator configuredActuator = configurationClass
                                .newInstance();
                        configuredActuator.setKey(actuator.getKey());
                        configuredActuator.setActuatorClass(actuator
                                .getActuatorClass());
                        configuredActuator.parseOptions(actuator
                                .getActuatorOptions());

                        initializeReconnectableXcfTypes(configuredActuator);

                        configuredActuatorsByKey.put(actuator.getKey(),
                                configuredActuator);

                        // stop searching for this actuator
                        continue actuatorLoop;

                    } else {
                        logger.debug("Configuration class "
                                + configurationClass
                                + " cannot create actuator " + actuator);
                    }

                } catch (NoSuchMethodException e) {
                    // CHECKSTYLE:OFF
                    assert false : "A method 'isSuitableFor' "
                            + "must exist for each configuration class.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, actuator.getKey());
                } catch (IllegalAccessException e) {
                    // CHECKSTYLE:OFF
                    assert false : "A method 'isSuitableFor' "
                            + "must exist for each configuration class "
                            + "and must be accessible.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, actuator.getKey());
                } catch (InvocationTargetException e) {
                    // CHECKSTYLE:OFF
                    assert false : "A method 'isSuitableFor' "
                            + "must exist for each configuration class "
                            + "and must work without exceptions.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, actuator.getKey());
                } catch (InstantiationException e) {
                    // CHECKSTYLE:OFF
                    assert false : "Configuration objects must be "
                            + "default constructible.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, actuator.getKey());
                }

            }

            logger.error("Error while configuring " + actuator.getKey()
                    + "! Implementation for Actuator "
                    + actuator.getActuatorClass() + " is unknown. ");

        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void configureWorkingMemories(Set<WorkingMemoryToConfigure> memories)
            throws IllegalArgumentException, CoreObjectCreationException {

        logger.info("Configuring memories: " + memories);

        memoryLoop:
        for (WorkingMemoryToConfigure memory : memories) {
            logger.debug("Processing working memory to configure: " + memory);

            // check that there is no other actuator with the requested key
            if (configuredMemoriesByKey.containsKey(memory.getKey())) {
                throw new IllegalArgumentException(
                        "Duplicate working memory key '" + memory.getKey()
                        + "' requested.");
            }

            // find the memory configuration that can handle this requested
            // memory
            for (Class<? extends ConfiguredXcfWorkingMemory> configurationClass : knownWorkingMemories) {

                logger.debug("Checking if configuration class "
                        + configurationClass + " can create working memory "
                        + memory);

                try {

                    Method isSuitableForMethod = configurationClass.getMethod(
                            IS_SUITABLE_METHOD_NAME,
                            WorkingMemoryToConfigure.class);
                    Boolean isSuitable = (Boolean) isSuitableForMethod.invoke(
                            null, memory);

                    // if this configuration object is suitable, create an
                    // instance and initialize it
                    if (isSuitable) {

                        logger.debug("Configuration class "
                                + configurationClass + " can create working memory "
                                + memory);

                        ConfiguredXcfWorkingMemory configuredMemory = configurationClass
                                .newInstance();
                        configuredMemory.setKey(memory.getKey());
                        configuredMemory.setWorkingMemoryClass(memory
                                .getMemoryClass());
                        configuredMemory
                                .parseOptions(memory.getMemoryOptions());
                        configuredMemory.configureMemorySlots(memory.getSlots());
                        initializeReconnectableXcfTypes(configuredMemory);

                        configuredMemoriesByKey.put(memory.getKey(),
                                configuredMemory);

                        // stop searching for this working memory
                        continue memoryLoop;

                    } else {
                        logger.debug("Configuration class "
                                + configurationClass
                                + " cannot create working memory " + memory);
                    }

                } catch (NoSuchMethodException e) {
                    // CHECKSTYLE:OFF
                    assert false : "A method 'isSuitableFor' "
                            + "must exist for each configuration class.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, memory.getKey());
                } catch (IllegalAccessException e) {
                    // CHECKSTYLE:OFF
                    assert false : "A method 'isSuitableFor' "
                            + "must exist for each configuration class "
                            + "and must be accessible.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, memory.getKey());
                } catch (InvocationTargetException e) {
                    // CHECKSTYLE:OFF
                    assert false : "A method 'isSuitableFor' "
                            + "must exist for each configuration class "
                            + "and must work without exceptions.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, memory.getKey());
                } catch (InstantiationException e) {
                    // CHECKSTYLE:OFF
                    assert false : "Configuration objects must be "
                            + "default constructible.";
                    // CHECKSTYLE:ON
                    throw new CoreObjectCreationException(e, memory.getKey());
                }

            }

            logger.error("Error while configuring " + memory.getKey()
                    + "! Implementation for working memory "
                    + memory.getMemoryClass() + " is unknown. ");

        }
    }

    private void initializeReconnectableXcfTypes(
            ConfiguredXcfCoreObject configuredActuator) {

        try {
            remoteServerManager
                    .initializeXcfTypes(new HashSet<String>(Arrays
                    .asList(configuredActuator
                    .getRemoteServersToInitialize())));
            subscriberManager.initializeXcfTypes(new HashSet<String>(Arrays
                    .asList(configuredActuator.getSubscribersToInitialize())));

            // TODO manage active memories when this is possible

            // CHECKSTYLE:OFF be careful here. Every error here means that the
            // core object could not be created.
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new CoreObjectCreationException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings({"unchecked", "rawtypes"})
    @Override
    public void configureSensors(Set<SensorToConfigure> sensors)
            throws IllegalArgumentException, CoreObjectCreationException {

        logger.info("Configuring sensors: " + sensors);

        // iterate over all sensors to configure
        sensorLoop:
        for (SensorToConfigure sensor : sensors) {
            // for (SensorToConfigure sensor : sensors) {

            logger.debug("Processing sensor configuration: " + sensor);

            // check that there is no other sensor with the requested key
            if (configuredSensorsByKey.containsKey(sensor.getKey())) {
                throw new IllegalArgumentException("Duplicate sensor key '"
                        + sensor.getKey() + "' requested.");
            }

            if (!sensor.isListSensor()) {
                // find the sensor configuration that can handle this requested
                // actuator
                for (Class<? extends ConfiguredXcfSensor> configurationClass : knownSensors) {

                    logger.debug("Checking if configuration class "
                            + configurationClass + " can create sensor "
                            + sensor);

                    try {

                        Method isSuitableForMethod = configurationClass
                                .getMethod(IS_SUITABLE_METHOD_NAME,
                                SensorToConfigure.class);
                        Boolean isSuitable = (Boolean) isSuitableForMethod
                                .invoke(null, sensor);

                        // if this configuration object is suitable, create an
                        // instance and initialize it
                        if (isSuitable) {

                            logger.debug("Configuration class "
                                    + configurationClass
                                    + " can create sensor " + sensor);

                            ConfiguredXcfSensor<?> configuredSensor = configurationClass
                                    .newInstance();
                            configuredSensor.setKey(sensor.getKey());
                            configuredSensor.setDataType((Class) sensor
                                    .getDataTypeClass());
                            configuredSensor.parseOptions(sensor
                                    .getSensorOptions());

                            initializeReconnectableXcfTypes(configuredSensor);

                            configuredSensorsByKey.put(sensor.getKey(),
                                    configuredSensor);

                            // stop searching for this sensor
                            continue sensorLoop;

                        } else {
                            logger.debug("Configuration class "
                                    + configurationClass
                                    + " cannot create sensor " + sensor);
                        }

                    } catch (NoSuchMethodException e) {
                        assert false : "A method 'isSuitableFor' "
                                + "must exist for each configuration class.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (IllegalAccessException e) {
                        assert false : "A method 'isSuitableFor' "
                                + "must exist for each configuration class "
                                + "and must be accessible.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (InvocationTargetException e) {
                        assert false : "A method 'isSuitableFor' "
                                + "must exist for each configuration class "
                                + "and must work without exceptions.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (InstantiationException e) {
                        assert false : "Configuration objects must be "
                                + "default constructible.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (ClassCastException e) {
                        assert false : "isSuitableFor must check "
                                + "that the data type is accepted.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    }
                }
            } else {

                // find the sensor configuration that can handle this requested
                // actuator
                for (Class<? extends ConfiguredXcfSensorListable> configurationClass : knownListSensors) {

                    logger.debug("Checking if configuration class "
                            + configurationClass + " can create sensor "
                            + sensor);

                    try {

                        Method isSuitableForMethod = configurationClass
                                .getMethod(IS_SUITABLE_METHOD_NAME,
                                SensorToConfigure.class);
                        Boolean isSuitable = (Boolean) isSuitableForMethod
                                .invoke(null, sensor);

                        // if this configuration object is suitable, create an
                        // instance and initialize it
                        if (isSuitable) {

                            logger.debug("Configuration class "
                                    + configurationClass
                                    + " can create list sensor " + sensor);

                            ConfiguredXcfSensorListable<List<?>, ?> configuredSensor = configurationClass
                                    .newInstance();
                            configuredSensor.setKey(sensor.getKey());
                            configuredSensor.setDataType((Class) sensor
                                    .getDataTypeClass());
                            configuredSensor.setListType((Class) sensor
                                    .getListTypeClass());
                            configuredSensor.parseOptions(sensor
                                    .getSensorOptions());

                            initializeReconnectableXcfTypes(configuredSensor);

                            configuredListSensorsByKey.put(sensor.getKey(),
                                    configuredSensor);

                            // stop searching for this sensor
                            continue sensorLoop;

                        } else {
                            logger.debug("Configuration class "
                                    + configurationClass
                                    + " cannot create sensor " + sensor);
                        }

                    } catch (NoSuchMethodException e) {
                        assert false : "A method 'isSuitableFor' "
                                + "must exist for each configuration class.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (IllegalAccessException e) {
                        assert false : "A method 'isSuitableFor' "
                                + "must exist for each configuration class "
                                + "and must be accessible.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (InvocationTargetException e) {
                        assert false : "A method 'isSuitableFor' "
                                + "must exist for each configuration class "
                                + "and must work without exceptions.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (InstantiationException e) {
                        assert false : "Configuration objects must be "
                                + "default constructible.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    } catch (ClassCastException e) {
                        assert false : "isSuitableFor must check "
                                + "that the data type is accepted.";
                        throw new CoreObjectCreationException(e, sensor.getKey());
                    }

                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    @Override
    public <T extends Actuator> T createActuator(String key,
            Class<T> actuatorClass) throws IllegalArgumentException,
            CoreObjectCreationException {

        // first check that the requested actuator can be created
        if (!canCreateActuator(key, actuatorClass)) {
            throw new IllegalArgumentException("No actuator with key '" + key
                    + "' and interface class '" + actuatorClass
                    + "' can be created by this factory.");
        }

        ConfiguredXcfActuator configuration = configuredActuatorsByKey.get(key);
        Actuator actuator = configuration.createInstance();

        initializeXcfObjectInstance(actuator);

        try {
            return (T) actuator;
        } catch (ClassCastException e) {
            assert false : "canCreateActuator seems to be wrong...";

            throw new IllegalArgumentException("No actuator with key '" + key
                    + "' and interface class '" + actuatorClass
                    + "' can be created by this factory.");
        }

    }

    private void initializeXcfObjectInstance(Object objectToInitialize) {

        if (objectToInitialize instanceof RemoteServerReconnectable) {
            remoteServerManager
                    .initializeAndManage((RemoteServerReconnectable) objectToInitialize);
        }
        if (objectToInitialize instanceof SubscriberReconnectable) {
            subscriberManager
                    .initializeAndManage((SubscriberReconnectable) objectToInitialize);
        }
        // TODO dummy code until it is clear how active memories can be
        // reconnected (see class todo)
        if (objectToInitialize instanceof ActiveMemoryReconnectable) {
            try {
                ActiveMemoryReconnectable reconnectable = (ActiveMemoryReconnectable) objectToInitialize;
                reconnectable
                        .resetActiveMemory(xcfManager
                        .createActiveMemory(reconnectable
                        .getActiveMemoryName()));
                // CHECKSTYLE:OFF catch every possible error while creating an
                // xcf connection
            } catch (Exception e) {
                // CHECKSTYLE:ON
                throw new XcfTypeInitializationException(e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    @Override
    public <T> Sensor<T> createSensor(String key, Class<T> dataType)
            throws IllegalArgumentException, CoreObjectCreationException {

        // first check that the requested sensor can be created
        if (!canCreateSensor(key, dataType)) {
            throw new IllegalArgumentException("No sensor with key '" + key
                    + "' and dataType '" + dataType
                    + "' can be created by this factory.");
        }

        ConfiguredXcfSensor<?> configuration = configuredSensorsByKey.get(key);
        Sensor<?> sensor = configuration.createInstance();

        initializeXcfObjectInstance(sensor);

        try {
            return (Sensor<T>) sensor;
        } catch (ClassCastException e) {
            assert false : "canCreateSensor seems to be wrong...";

            throw new IllegalArgumentException("No sensor with key '" + key
                    + "' and data type class '" + dataType
                    + "' can be created by this factory.");
        }

    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    @Override
    public <S extends List<T>, T> Sensor<S> createSensor(String key,
            Class<S> listType, Class<T> dataType)
            throws IllegalArgumentException, CoreObjectCreationException {

        // first check that the requested sensor can be created
        if (!canCreateSensor(key, listType, dataType)) {
            throw new IllegalArgumentException("No list sensor with key '"
                    + key + "', listType '" + listType + "' and dataType '"
                    + dataType + "' can be created by this factory.");
        }

        ConfiguredXcfSensorListable<S, T> configuration = (ConfiguredXcfSensorListable<S, T>) configuredListSensorsByKey
                .get(key);
        Sensor<S> sensor = configuration.createInstanceListable();

        initializeXcfObjectInstance(sensor);

        try {
            return sensor;
        } catch (ClassCastException e) {
            assert false : "canCreateSensor seems to be wrong...";

            throw new IllegalArgumentException("No sensor with key '" + key
                    + "', list type class '" + listType
                    + "' and data type class '" + dataType
                    + "' can be created by this factory.");
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void initialize(Map<String, String> options)
            throws IllegalArgumentException, InitializationException {

        // setup xcf connection
        try {

            xcfManager = XcfManager.createXcfManager();

            // subscriber manager
            if (options.containsKey(OPTION_SUBSCRIBER_CHECK_INTERVAL)) {
                subscriberManager = new SubscriberReconnectableManager(
                        xcfManager, Long.parseLong(options
                        .get(OPTION_SUBSCRIBER_CHECK_INTERVAL)));
            } else {
                subscriberManager = new SubscriberReconnectableManager(
                        xcfManager);
            }

            // remote server manager
            if (options.containsKey(OPTION_REMOTE_SERVER_CHECK_INTERVAL)) {
                remoteServerManager = new RemoteServerReconnectableManager(
                        xcfManager, Long.parseLong(options
                        .get(OPTION_REMOTE_SERVER_CHECK_INTERVAL)));
            } else {
                remoteServerManager = new RemoteServerReconnectableManager(
                        xcfManager);
            }

            // errors on initial subscription
            if (options.containsKey(OPTION_ERROR_ON_INITIAL_SUBSCRIPTION)) {
                subscriberManager.setErrorOnXcfTypeInitialization(Boolean
                        .parseBoolean(options
                        .get(OPTION_ERROR_ON_INITIAL_SUBSCRIPTION)));
            } else {
                subscriberManager
                        .setErrorOnXcfTypeInitialization(ReconnectableManager.DEFAULT_ERROR_ON_TYPE_INITIALIZATION);
            }

            // discover sensors and actuators
            discoverSensorsAndActuators();

        } catch (InitializeException e) {
            throw new InitializationException("Error connecting to XCF.", e);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException("Error parsing a number.", e);
        }

    }

    @SuppressWarnings({"rawtypes", "unchecked"})
    protected void discoverSensorsAndActuators() {
        knownActuators = serviceDiscovery
                .discoverServicesByInterface(ConfiguredXcfActuator.class);
        knownSensors = serviceDiscovery
                .discoverServicesByInterface(ConfiguredXcfSensor.class);
        knownWorkingMemories = serviceDiscovery
                .discoverServicesByInterface(ConfiguredXcfWorkingMemory.class);

        // search for subclasses of ConfiguredXcfSensorListable
        for (Class<? extends ConfiguredXcfSensor> clazz : knownSensors) {
            if (ConfiguredXcfSensorListable.class.isAssignableFrom(clazz)) {
                knownListSensors
                        .add((Class<? extends ConfiguredXcfSensorListable>) clazz);
            }
        }
    }

    @Override
    public <T extends WorkingMemory> T createWorkingMemory(String key)
            throws IllegalArgumentException, CoreObjectCreationException {
        // first check that the requested actuator can be created
        if (!canCreateWorkingMemory(key)) {
            throw new IllegalArgumentException("No working memory with key '" + 
                    key + "' can be created by this factory.");
        }

        ConfiguredXcfWorkingMemory configuration = configuredMemoriesByKey.get(key);
        WorkingMemory memory = configuration.createInstance();

        initializeXcfObjectInstance(memory);

        try {
            return (T) memory;
        } catch (ClassCastException e) {
            assert false : "canCreateWorkingMemory seems to be wrong...";

            throw new IllegalArgumentException("No working memory with key '" + 
                    key + "' can be created by this factory.");
        }
    }

    @Override
    public boolean canCreateWorkingMemory(String key) {
        return configuredMemoriesByKey.containsKey(key)
                && WorkingMemory.class.isAssignableFrom(
                configuredMemoriesByKey.get(key).getWorkingMemoryClass());
    }
}
