package de.unibi.citec.clf.bonsai.xcf.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.actuators.ROSDynamicReconfigurationActuator;
import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.navigation.ROSDynamicConfiguration;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import nu.xom.Element;
import org.apache.log4j.Logger;

/**
 * An actuator which is capable of either writing a configuration to dynamically 
 * change ROS parameter of reconfigurable ROS nodes, or reading its current 
 * configuration. There are several different configurations which may be 
 * applied.
 * 
 * @author pdressel
 */
public class XcfROSDynamicReconfigurationActuator<T extends ROSDynamicConfiguration> 
        implements ROSDynamicReconfigurationActuator<T>, RemoteServerReconnectable 
{ 
    private Logger logger = Logger.getLogger(getClass());
   
    private String remoteServerName = "";
    private String serverMethodNameRead = "";
    private String serverMethodNameWrite = "";
    
    private Object remoteServerLock = new Object();
    private RemoteServer remoteServer = null;
     
    Class<T> typeClass = null;
    private Method fromDocumentMethod = null;
    private Method getBaseTag = null;
    
    /**
     * Constructor.
     * 
     * @param remoteServerName
     *      Name of the remote XcfServer to connect to.
     * @param serverMethodNameRead
     *      Name of the advertised "read method".
     * @param serverMethodNameWrite 
     *      Name of the advertised "write method".
     */
    public XcfROSDynamicReconfigurationActuator(
            Class<T> typeClass, 
            String remoteServerName, 
            String serverMethodNameRead,
            String serverMethodNameWrite)
    {
        logger.debug("Creating ROSDynamicReconfigureActuator.");
        
        this.typeClass = typeClass;
        this.remoteServerName = remoteServerName;

        this.serverMethodNameRead = serverMethodNameRead;
        this.serverMethodNameWrite = serverMethodNameWrite;
        
        try {            
            // Get the "fromDocument" and "getBaseTag" method from an object of type T via 
            // reflection
            fromDocumentMethod = typeClass.getMethod(
                    "fromDocument", 
                    new Class[] { Document.class, Class.class });
            
            getBaseTag = typeClass.getMethod(
                    "getBaseTag",
                    new Class[] { });
            
        } catch (SecurityException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        } 
    }

    /**
    * {@inheritDoc}
    */
    @Override
    public void writeConfig(T config) throws IOException {
        
        try {
            logger.debug("WriteConfig called: Preparing Document (xml)");

            Document doc = new Document(new Element((String) getBaseTag.invoke(null)));
            config.fillInto(doc.getRootElement());

            logger.debug("Calling " + serverMethodNameWrite 
                    + " advertised by " + remoteServerName 
                    + " server with " + doc.toXML());

            XOPData xop;
            synchronized (remoteServerLock) {     
                remoteServer.callMethod(serverMethodNameWrite, new XOPData(doc));
            }

            logger.debug("Call completed!\n");

        } catch (Exception e) {
            throw new IOException(e);
        }
    }
    
    @Override
    public T readConfig() throws IOException {
        
        try {
            logger.debug("Calling " + serverMethodNameRead 
                + " advertised by " + remoteServerName 
                + " server");
            
            XOPData xop;
            synchronized (remoteServerLock) {
                xop = remoteServer.callMethod(serverMethodNameRead, new XOPData());
            }

            logger.debug("Call completed. Received message: " + xop.getDocument().toXML());            
            
            T typeData = safeExtractType(xop);           
            return typeData;        
            
        } catch (Exception e) {
            throw new IOException(e);
        }
    }  
   
    /**
     * Safely extracts the BTL type from the document found in the
     * {@link XOPData}.
     * 
     * @param data
     *            data containing the document
     * @return extracted BTL type
     * @throws IOException
     *             error extracting BTL type instance
     */    
    protected T safeExtractType(XOPData data)
            throws IOException {
        
        logger.info("Parsing event to configuration");

        Document doc = data.getDocument();
        if (doc == null)
            throw new IOException("Document empty in XOPData: " + data.getDocumentAsText());
        
        try {
            
            Object rawResult = fromDocumentMethod.invoke(null, doc, typeClass);
            if (rawResult == null) {
                throw new IOException(
                        "fromDocument method has wrong signature "
                                + "or simply returned null.");
            }
            
            postProcessType(data, (T) rawResult);
            return (T) rawResult;
            
        } catch (IllegalArgumentException e) {
            throw new IOException(e);
        } catch (IllegalAccessException e) {
            throw new IOException(e);
        } catch (InvocationTargetException e) {
            throw new IOException("Error parsing the document.", e);
        } catch (ExceptionInInitializerError e) {
            throw new IOException(e);
        }
    }    
    
    /**
     * Hook method that may be implemented in subclasses to post process the
     * simple BTL based de-serialization of the data type.
     * 
     * @param data
     *            {@link XOPData} for the event
     * @param extractedType
     *            type already extract with BTL <code>fromDocument</code>
     * @throws IOException
     *             the post processing failed and returning the type to callers
     *             would be illegal
     */
    protected void postProcessType(XOPData data, T extractedType)
            throws IOException {

    }       
    
    /**
     * {@inheritDoc}
     */    
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;        
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        
        assert remoteServer.getServerName().equals(remoteServerName);
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }
    }
}
