package de.unibi.citec.clf.bonsai.xcf.configuration;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.Actuator;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.ActuatorToConfigure;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.CoreObjectCreationException;
import java.util.Map;

/**
 * Base class for configuration objects that create actuators. For convenience
 * reasons the default implementation of the {@link #parseOptions(Map)} method
 * already parses a required server name.
 * 
 * @author jwienke
 */
public abstract class ConfiguredXcfActuator extends ConfiguredXcfCoreObject {

    /**
     * Option for remote server based classes defining the remote server name.
     */
    public static final String OPTION_SERVER_NAME = "serverName";

    /**
     * Name of the server if parsed by {@link #parseOptions(Map)}.
     */
    private String serverName = "";

    /**
     * Concrete class for the actuator requested by the user in the
     * configuration.
     */
    private Class<? extends Actuator> actuatorClass = null;

    /**
     * Override this!
     * 
     * @param actuatorToConfigure
     *            requested configuration
     * @return <code>true</code> if this configuration object is suitable for
     *         the requested actuator
     */
    public static boolean isSuitableFor(ActuatorToConfigure actuatorToConfigure) {
        return false;
    }

    /**
     * Create a new specific actuator instance.
     * 
     * @return new instance
     * @throws CoreObjectCreationException
     *             problem creating the new instance
     */
    public abstract Actuator createInstance()
            throws CoreObjectCreationException;

    /**
     * {@inheritDoc}
     * 
     * Default implementation parsing the server name.
     */
    @Override
    public void parseOptions(Map<String, String> options)
            throws IllegalArgumentException {

        if (!options.containsKey(OPTION_SERVER_NAME)) {
            throw new IllegalArgumentException(
                    "Server name missing in arguments for requested type");
        }

        this.serverName = options.get(OPTION_SERVER_NAME);

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String[] getRemoteServersToInitialize() {
        return new String[] { getServerName() };
    }

    /**
     * Returns the configured server name if this class'
     * {@link #parseOptions(Map)} method is used.
     * 
     * @return configured server name
     */
    public String getServerName() {
        return serverName;
    }

    /**
     * Setter for the implementation class requested by the user.
     * 
     * @param actuatorClass
     *            new actuator class
     */
    public void setActuatorClass(Class<? extends Actuator> actuatorClass) {
        this.actuatorClass = actuatorClass;
    }

    /**
     * Returns the concrete implementation class for the actuator requested in
     * the configuration.
     * 
     * @return implementation class for the actuator
     */
    public Class<? extends Actuator> getActuatorClass() {
        return actuatorClass;
    }

}
