package de.unibi.citec.clf.bonsai.xcf.configuration;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.Actuator;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.ActuatorToConfigure;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.CoreObjectCreationException;
import de.unibi.citec.clf.bonsai.xcf.actuators.XcfROSDynamicReconfigurationActuator;
import de.unibi.airobots.btl.data.navigation.ROSDynamicConfiguration;
import java.util.Map;

/**
 * Configuration object for {@link XcfROSDynamicReconfigurationActuator}
 *  
  * <p>
 * Available options:
 * <ul>
 * <li>{@link XcfROSDynamicReconfigurationActuator#OPTION_MEMORY_NAME} (required): name of the
 * server to register to</li>
 * <li>{@link XcfROSDynamicReconfigurationActuator#OPTION_SERVERMETHODNAME_APPLY} (required): name of the 
 * method of the registered server to send configurations to</li>
 * </ul>
 * 
 * @author pdressel
 */
public class ConfiguredXcfROSDynamicReconfigurationActuator<T extends ROSDynamicConfiguration> 
        extends ConfiguredXcfActuator {    
    
    /**
     * Expected tag name for defining the data type which is read and written by
     * this actuator.
     */
    public static final String OPTION_DATATYPECLASS = "dataTypeClass";
    
    /**
     * Expected tag name for defining the method name of the "read method" 
     * advertised by the corresponding XcfServer in the actuator configuration 
     * file. 
     */
    public static final String OPTION_SERVERMETHODNAME_READ = "methodNameRead";
    
    /**
     * Expected tag name for defining the method name of the "write method" 
     * advertised by the corresponding XcfServer in the actuator configuration 
     * file. 
     */    
    public static final String OPTION_SERVERMETHODNAME_WRITE = "methodNameWrite";
    
    /**
     * The name of the method of the XcfServer to call to read a configuration
     * of type T.
     */
    private String serverMethodNameRead = "";
    
    /**
     * The name of the method of the XcfServer to call to write a configuration
     * of type T.
     */    
    private String serverMethodNameWrite = "";
     
    /**
     * The class of the objects this actuator may use.
     */
    private Class<T> dataTypeClass = null;
    
    /**
     * {@inheritDoc}
     */
    @Override
    public Actuator createInstance() throws CoreObjectCreationException {
        return new XcfROSDynamicReconfigurationActuator<T>(
                getDataTypeClass(),
                getServerName(), 
                serverMethodNameRead,
                serverMethodNameWrite);
    }
        
    /**
     * @return the serverMethodNameRead
     */
    public String getServerMethodNameRead() {
        return serverMethodNameRead;
    }

    /**
     * @return the serverMethodNameWrite
     */
    public String getServerMethodNameWrite() {
        return serverMethodNameWrite;
    }

    /**
     * @return the dataTypeClass
     */
    public Class<T> getDataTypeClass() {
        return dataTypeClass;
    }
    
    /**
     * Override this and check for suitable data type!
     * 
     * @param sensor
     *            requested sensor
     * @return <code>true</code> if configuration can create such a requested
     *         sensor
     */    
    public static boolean isSuitableFor(ActuatorToConfigure actuatorToConfigure)
    {
        return actuatorToConfigure.getActuatorClass().equals(
                XcfROSDynamicReconfigurationActuator.class);
    }
    
    @Override
    public void parseOptions(Map<String, String> options)
            throws IllegalArgumentException {

    	super.parseOptions(options);
    	
        if (!options.containsKey(OPTION_DATATYPECLASS)) {
            throw new IllegalArgumentException("Missing option with name: " + OPTION_DATATYPECLASS);
        } else {
            try {
                this.dataTypeClass = (Class<T>) Class.forName(options.get(OPTION_DATATYPECLASS));
                System.out.println("Classname: " + this.dataTypeClass.getName());
            } catch (ClassNotFoundException ex) {
                throw new IllegalArgumentException("Invalid option with name: " + OPTION_DATATYPECLASS);
            }
        }
        
        if (!options.containsKey(OPTION_SERVERMETHODNAME_READ)) {
            throw new IllegalArgumentException("Missing option with name: " + OPTION_SERVERMETHODNAME_READ);
        } else {            
            this.serverMethodNameRead = options.get(OPTION_SERVERMETHODNAME_READ);        
        }
        
        // If the options are not present the default value remains
        if (!options.containsKey(OPTION_SERVERMETHODNAME_WRITE)) {
            throw new IllegalArgumentException("Missing option with name: " + OPTION_SERVERMETHODNAME_WRITE);
        } else {
            this.serverMethodNameWrite = options.get(OPTION_SERVERMETHODNAME_WRITE);
        }
    }           
}