package de.unibi.citec.clf.bonsai.xcf.sensors;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.event.MemoryEvent;
import net.sf.xcf.event.MemoryEventAdapter;
import net.sf.xcf.event.MemoryEventListener;
import net.sf.xcf.memory.MemoryAction;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.xml.XPath;
import nu.xom.ParsingException;

import org.apache.log4j.Logger;

import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.person.PersonData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;
import de.unibi.citec.clf.bonsai.core.Sensor;
import de.unibi.citec.clf.bonsai.core.SensorListener;
import de.unibi.citec.clf.bonsai.sensors.PersonSensor;
import de.unibi.citec.clf.bonsai.sensors.data.PersonList;
import de.unibi.citec.clf.bonsai.util.CoordinateSystemConverter;
import de.unibi.citec.clf.bonsai.xcf.ActiveMemoryReconnectable;

/**
 * The person sensor receives the person list provided by the PTA (MMAnchoring).
 * The sensor provides the list with all persons currently in the robots memory.
 * 
 * @author lschilli
 * @author jwienke
 */
public class XcfActiveMemoryPersonSensor implements ActiveMemoryReconnectable,
		PersonSensor {

	private final static long MIN_TIME_DIFFERENCE_MS = 500;
	private final static double MIN_DISTANCE_M = 0.1;

	private Logger logger = Logger.getLogger(getClass());

	private final Object sharedDataLock = new Object();

	private final Object personMapLock = new Object();
	private final Object sampledPersonMapLock = new Object();

	private Sensor<PositionData> positionSensor;

	/**
	 * Map of {@link PersonData} indexed by their id. Always acquire
	 * {@link #sharedDataLock} before using this map.
	 */
	private HashMap<Integer, PersonData> personmap = new HashMap<Integer, PersonData>();

	private HashMap<Integer, PersonData> sampledPersonmap = new HashMap<Integer, PersonData>();

	/**
	 * Own listener used to intercept the memory for person changes.
	 */
	private MemoryEventListener selfListener;

	/**
	 * Set of sensor listeners. Only access this set after acquiring
	 * {@link #sharedDataLock}.
	 */
	private Set<SensorListener<PersonList>> listeners = new HashSet<SensorListener<PersonList>>();

	private String memoryName = "";
	private final Object memoryLock = new Object();
	private ActiveMemory memory = null;

	/**
	 * Constructor.
	 * 
	 * @param memoryName
	 *            name of the active memory to get the persons from
	 */
	public XcfActiveMemoryPersonSensor(String memoryName) {

		this.memoryName = memoryName;

		logger.debug("Create memory person Sensor for memoryName " + memoryName);

		selfListener = new MemoryEventAdapter(MemoryAction.ALL, new XPath(
				"/PERSON")) {

			@Override
			public void handleEvent(MemoryEvent e) {
				try {

					if (e.getType() == MemoryAction.INSERT
							|| e.getType() == MemoryAction.REPLACE) {

						// Extract the person data from the xcf
						PersonData p = PersonData.fromDocument(e.getData()
								.getDocument(), PersonData.class);

						// Remove persons with invalid id's
						if (p.getId() < 0) {
							return;
						}

						PositionData robotPosition = null;
						if (positionSensor != null) {
							robotPosition = positionSensor.readLast();
						}

						if (robotPosition != null) {
							PositionData pPosition = CoordinateSystemConverter
									.polar2NavigationGoalData(robotPosition,
											p.getAngle(AngleUnit.RADIAN),
											p.getDistance(LengthUnit.METER),
											AngleUnit.RADIAN, LengthUnit.METER);

							// A person is either added to the sampledPersonmap
							// if
							// no person with the same id is in the list yet or
							// its
							// timestamp is older than MIN_TIME_DIFFERENCE
							synchronized (sampledPersonMapLock) {
								if (!sampledPersonmap.containsKey(p.getId())) {
									sampledPersonmap.put(p.getId(), p);
								} else {

									PersonData lastSampledPerson = sampledPersonmap
											.get(p.getId());
									PositionData lastSampledPersonPosition = CoordinateSystemConverter
											.polar2NavigationGoalData(
													robotPosition,
													lastSampledPerson
															.getAngle(AngleUnit.RADIAN),
													lastSampledPerson
															.getDistance(LengthUnit.METER),
													AngleUnit.RADIAN,
													LengthUnit.METER);

									double distanceMeter = pPosition
											.getDistance(
													lastSampledPersonPosition,
													LengthUnit.METER);
									double timeDifference = p.getTimestamp()
											.getUpdated().getTime()
											- lastSampledPerson.getTimestamp()
													.getUpdated().getTime();

									// logger.error("Person distance (" +
									// p.getId() + "):" + distanceMeter);
									// If the distance to the lastSampledPerson
									// is
									// great enough the person is considered
									// walking
									if (distanceMeter > MIN_DISTANCE_M)
										p.setWalking(true);
									else
										p.setWalking(false);

									// If the the time difference to the
									// lastSampledPerson
									// is great enough it is replaced by p
									if (timeDifference > MIN_TIME_DIFFERENCE_MS)
										sampledPersonmap.put(p.getId(), p);
								}
							}
						}

						// Add the person to the intern list and inform all
						// listeners.
						synchronized (personMapLock) {
							// logger.error("Person isWalking: " +
							// p.isWalking());
							personmap.put(p.getId(), p);
						}
                                                publish();

					} else if (e.getType() == MemoryAction.REMOVE) {

						// Extract the person data from the xcf
						PersonData p = PersonData.fromDocument(e.getData()
								.getDocument(), PersonData.class);

						// Remove the person from the intern list an infoem all
						// listeners
						synchronized (personMapLock) {
							personmap.remove(p.getId());
						}
                                                publish();
					}
				} catch (InterruptedException ex) {
					logger.error(ex);
				} catch (IOException ex) {
					logger.error(ex);
				} catch (ParsingException ex) {
					logger.error(ex);
				}
			}
		};

		logger.debug("End memory person Sensor for memoryName ");

	}
	
	@Override
	public void setPositionSensor(Sensor<PositionData> posSens) {
		positionSensor = posSens;
	}

	private PersonList getPersonListFromMemory() throws IOException {
		PersonList pl = new PersonList();
		synchronized (personMapLock) {

			for (PersonData p : personmap.values()) {
				pl.addPerson(p);
			}
		}

		return pl;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void addSensorListener(SensorListener<PersonList> listener) {
		synchronized (sharedDataLock) {
			listeners.add(listener);
		}
	}

	@Override
	public PersonList readMovingPersons() {

		PersonList movingList = new PersonList();

		synchronized (personMapLock) {
			for (PersonData person : this.personmap.values()) {
				if (person.isWalking())
					movingList.addPerson(person);
			}
		}

		return movingList;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Class<PersonList> getDataType() {
		return PersonList.class;
	}

	/**
	 * Delivers a list of the last persons tracked by the sensor.
	 * 
	 * {@inheritDoc}
	 */
	@Override
	public PersonList readLast() throws IOException {
		return getPersonListFromMemory();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public PersonList readLast(long timeout) throws IOException,
			InterruptedException {
		return readLast();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void removeSensorListener(SensorListener<PersonList> listener) {
		synchronized (sharedDataLock) {
			listeners.remove(listener);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String getActiveMemoryName() {
		return memoryName;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void resetActiveMemory(ActiveMemory activeMemory) {
		synchronized (memoryLock) {
			this.memory = activeMemory;
		}

		try {
			memory.addListener(selfListener);
		} catch (MemoryException e) {
			logger.error("Error adding listeners to new memory instance", e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean hasNext() {
		synchronized (personMapLock) {
			return personmap.size() > 0;
		}
	}

	/**
	 * Informs all listeners that new data is available. Sends the current
	 * PersonList.
	 */
	private void publish() {

		try {
			PersonList list = getPersonListFromMemory();

			for (SensorListener<PersonList> l : listeners) {
				l.newDataAvailable(list);
			}
		} catch (IOException ex) {
			logger.error("Error notifying observers", ex);
		}
	}
}
