package de.unibi.citec.clf.bonsai.xcf.sensors;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.SyntaxException;
import nu.xom.Document;
import nu.xom.Element;
import de.unibi.citec.clf.bonsai.sensors.FaceIdentificationSensor2;
import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.person.FaceIdentificationList;

/**
 * A sensor for face identification results provided by caiwicat.
 * 
 * @author jwienke
 */
public class XcfFaceIdentificationSensor2 extends
		XcfBtlPublisherSensor<FaceIdentificationList> implements
		RemoteServerReconnectable, FaceIdentificationSensor2 {

	// TODO these method could also be made configurable in the configuration
	// object

	// server method names
	private static final String TRAIN_METHOD_NAME = "train";
	private static final String IDLE_METHOD_NAME = "idle";
	private static final String PREDICT_METHOD_NAME = "predict";
	private static final String DETECT_METHOD_NAME = "detect";
	private static final String CAPTURE_NAME = "capture";
	private static final String CAPTURE_NEXT_METHOD_NAME = "captureNext";
	private static final String CAPTURE_CURRENT_METHOD_NAME = "captureCurrent";
	private static final String SET_ID_METHOD_NAME = "setId";
	private static final String GET_ID_METHOD_NAME = "getId";

	private String controlServerName = "";
	private Object controlServerLock = new Object();
	/**
	 * Server used to control the learning process.
	 */
	private RemoteServer controlServer;

	/**
	 * Constructor.
	 * 
	 * @param resultSubscriberName
	 *            name of the subscriber that receives the face identification
	 *            results
	 * @param activationServerName
	 *            name of the server to activate or deactivate identification
	 * @param controlServerName
	 *            name of the server used to control the identification and
	 *            learning process
	 */
	public XcfFaceIdentificationSensor2(String resultSubscriberName,
			String controlServerName) {

		super(FaceIdentificationList.class, resultSubscriberName);

		this.controlServerName = controlServerName;

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
		Map<String, ErrorHandlingPolicy> servers = new HashMap<String, ErrorHandlingPolicy>();
		servers.put(controlServerName, ErrorHandlingPolicy.SELF_HEALING);
		return servers;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void resetRemoteServer(RemoteServer remoteServer) {
		if (remoteServer.getServerName().equals(controlServerName)) {
			synchronized (controlServerLock) {
				controlServer = remoteServer;
			}
		} else {
			assert false : "Unknown server with name "
					+ remoteServer.getServerName() + " received.";
		}
	}

	@Override
	public void trainNow() throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(TRAIN_METHOD_NAME,
						XOPData.fromString("<TRAIN/>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public void setCurrentId(int id) throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(SET_ID_METHOD_NAME,
						XOPData.fromString("<ID>" + id + "</ID>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public int getCurrentId() throws IOException {
		try {
			synchronized (controlServerLock) {
				XOPData data = controlServer.callMethod(GET_ID_METHOD_NAME,
						XOPData.fromString("<GETID/>"));
				Document doc = data.getDocument();
				Element idElem = doc.getRootElement();
				int id = Integer.parseInt(idElem.getValue());
				return id;
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public void capture(int id) throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(CAPTURE_NAME,
						XOPData.fromString("<ID>" + id + "</ID>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public void captureNext() throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(CAPTURE_NEXT_METHOD_NAME,
						XOPData.fromString("<CAPTURENEXT/>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public void captureCurrent() throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(CAPTURE_CURRENT_METHOD_NAME,
						XOPData.fromString("<CAPTURECURRENT/>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public void idle() throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(IDLE_METHOD_NAME,
						XOPData.fromString("<IDLE/>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public void predict() throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(PREDICT_METHOD_NAME,
						XOPData.fromString("<PREDICT/>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}

	@Override
	public void detect() throws IOException {
		try {
			synchronized (controlServerLock) {
				controlServer.callMethod(DETECT_METHOD_NAME,
						XOPData.fromString("<DETECT/>"));
			}
		} catch (MethodNotFoundException e) {
			throw new IOException(e);
		} catch (UserException e) {
			throw new IOException(e);
		} catch (SyntaxException e) {
			throw new IOException(e);
		}
	}
}
