package de.unibi.citec.clf.bonsai.xcf.sensors;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.Sensor;
import de.unibi.citec.clf.bonsai.core.SensorListener;
import de.unibi.citec.clf.bonsai.sensors.data.ImageData;
import de.unibi.citec.clf.bonsai.sensors.data.KinectImageData;
import de.unibi.citec.clf.bonsai.sensors.data.PlaneType;
import de.unibi.citec.clf.bonsai.xcf.BoundSynchronizedQueue;
import de.unibi.citec.clf.bonsai.xcf.SubscriberReconnectable;
import de.unibi.airobots.btl.data.vision3d.KinectImageData.ImageType;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import net.sf.xcf.Subscriber;
import net.sf.xcf.event.PublishEvent;
import net.sf.xcf.event.PublishEventAdapter;
import net.sf.xcf.event.PublishEventListener;
import net.sf.xcf.event.XcfEvent;
import net.sf.xcf.transport.Attachment;
import org.apache.log4j.Logger;

public class XcfKinectImageSensor implements Sensor<KinectImageData>,
        SubscriberReconnectable {

    private final Logger logger = Logger.getLogger(getClass());

    private static final int BYTE_COUNT_RGB = 921600;
    private static final int BYTE_COUNT_DEPTH = 614400;
    private static final int BYTE_COUNT_REAL_DEPTH = 3686400;

    private Object subscriberLock = new Object();
    private Object listenersLock = new Object();
    private String subscriberName;
    private Subscriber subscriber;

    private PublishEventListener listenerAdapter = new ListenerNotifier();

    private BoundSynchronizedQueue<PublishEvent> queue =
            new BoundSynchronizedQueue<PublishEvent>() {

                @Override
                public PublishEvent convertFromXcfEvent(XcfEvent e) {
                    // not used here
                    return null;
                }
            };

    private List<SensorListener<KinectImageData>> listeners =
            new LinkedList<SensorListener<KinectImageData>>();

    public XcfKinectImageSensor(String subscriberName) {
        logger.info("Creating subscriber to: '" + subscriberName + "'");
        this.subscriberName = subscriberName;
    }

    @Override
    public void addSensorListener(SensorListener<KinectImageData> listener) {
        synchronized (listenersLock) {
            if (!listeners.contains(listener))
                listeners.add(listener);
        }
    }

    @Override
    public void removeSensorListener(SensorListener<KinectImageData> listener) {
        synchronized (listenersLock) {
            listeners.remove(listener);
        }
    }

    @Override
    public Class<KinectImageData> getDataType() {
        return KinectImageData.class;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean hasNext() {
        return this.queue.hasNext();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public KinectImageData readLast() throws IOException, InterruptedException {
        try {
            return fromPublishEvent(queue.next());
        } catch (IllegalArgumentException e) {
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public KinectImageData readLast(long timeout) throws IOException,
            InterruptedException {
        try {
            PublishEvent event = queue.next(timeout);
            if (event != null) {
                return fromPublishEvent(event);
            }
            return null;
        } catch (IllegalArgumentException e) {
            throw new IOException(e);
        }
    }

    @Override
    public String getSubscriberName() {
        return subscriberName;
    }

    @Override
    public void resetSubscriber(Subscriber subscriber) {
        synchronized (subscriberLock) {
            if (this.subscriber != null) {
                this.subscriber.removeListener(listenerAdapter);
            }
            this.subscriber = subscriber;
            this.subscriber.addListener(listenerAdapter);
        }
    }

    /**
     * Convert a {@link PublishEvent} to {@link ImageData}.
     * 
     * @param event
     *            event to convert
     * @return converted {@link ImageData}
     * @throws IllegalArgumentException
     *             converting failed
     */
    private KinectImageData fromPublishEvent(PublishEvent event)
            throws IllegalArgumentException {
        logger.info("Parsing Kinect image");
        try {
            de.unibi.airobots.btl.data.vision3d.KinectImageData btlData = new de.unibi.airobots.btl.data.vision3d.KinectImageData();
            de.unibi.airobots.btl.data.vision3d.KinectImageData.fromElement(event.getData().getDocument()
                    .getRootElement(), btlData);

            PlaneType ptype;
            int expectedSize;

            if (btlData.getType() == ImageType.RGB) {
                ptype = PlaneType.PLANAR_RGB8;
                expectedSize = BYTE_COUNT_RGB;
            } else if (btlData.getType() == ImageType.DEPTH) {
                ptype = PlaneType.GRAY16;
                expectedSize = BYTE_COUNT_DEPTH;
            } else {
                ptype = PlaneType.PLANAR_RGB32;
                expectedSize = BYTE_COUNT_REAL_DEPTH;
            }

            // get attachment
            Attachment imagePacket =
                    event.getData().getAttachment(btlData.getUri());

            if (imagePacket.getValue().length == expectedSize) {
                return new KinectImageData(btlData.getWidth(), btlData.getHeight(),
                        btlData.getUri(), imagePacket.getValue(), 0, ptype);
            } else {
                if (imagePacket.getValue().length > expectedSize){
                // old-school packet with meta-data header
                return new KinectImageData(btlData.getWidth(), btlData.getHeight(),
                        btlData.getUri(), imagePacket.getValue(), imagePacket
                                .getValue().length
                                - expectedSize, ptype);
                } else {
                    throw new IllegalArgumentException(
                            "Error parsing image from xcf event. Unexpected Attachment-size");
                }
            }
            // CHECKSTYLE:OFF catch everything here as error
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IllegalArgumentException(
                    "Error parsing image from xcf event", e);
        }
    }

    /**
     * An adapter that maps a {@link PublishEvent} to all registered
     * {@link SensorListener}s.
     * 
     * @author vrichter
     */
    private class ListenerNotifier extends PublishEventAdapter {

        @Override
        public void handleEvent(PublishEvent event) {

            queue.push(event);
            if (!listeners.isEmpty()) {

                KinectImageData image = fromPublishEvent(event);
                for (SensorListener<KinectImageData> l : listeners) {
                    l.newDataAvailable(image);
                }

            }

        }
    }

}
